'''Generic connection endpoint.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.


class Endpoint():
    '''Generic communication endpoint.'''
