'''Implementation of Ranker'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from .ranker import Ranker


class RankerImpl(Ranker):
    '''Real implementation of the Ranker'''
