'''Identifies nodes in a distributed AI setting'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.


class NodeID(int):
    '''Each neighbor is uniquely identified by an integer.'''
